@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2008-2012, Parallels, Inc. All rights reserved.
+ *  Copyright (C) 2008-2015, Parallels, Inc. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -107,7 +107,7 @@ static int grow_lower_delta(const char *device, int top, int start_level, int en
 	struct grow_maps grow_maps;
 	char *fmt;
 	int dst_is_raw = 0;
-	void *buf;
+	void *buf = NULL;
 	struct delta odelta = {};
 	int ret;
 	__u64 cluster;
@@ -172,11 +172,6 @@ static int grow_lower_delta(const char *device, int top, int start_level, int en
 	if ((ret = grow_delta(&odelta, src_size, buf, &grow_maps)))
 		goto done;
 
-	if (grow_maps.ctl->n_maps &&
-	    odelta.fops->update_size &&
-	    (ret = odelta.fops->update_size(odelta.fd, dst_image)))
-		goto done;
-
 	if (clear_delta(&odelta)) {
 		ploop_err(errno, "clear_delta");
 		ret = SYSEXIT_WRITE;
@@ -219,12 +214,13 @@ static int grow_lower_delta(const char *device, int top, int start_level, int en
 		goto done;
 	}
 
-	if (odelta.fops->fsync(odelta.fd)) {
+	if (fsync(odelta.fd)) {
 		ploop_err(errno, "fsync");
 		ret = SYSEXIT_FSYNC;
 	}
 
 done:
+	free(buf);
 	free(src_image);
 	free(dst_image);
 	close_delta(&odelta);
@@ -287,8 +283,23 @@ int get_delta_info(const char *device, struct merge_info *info)
 	return 0;
 }
 
+/* Merge one or more deltas (specified by **images and start/end_level).
+ * Merge can be online or offline, with or without top delta, to either
+ * a lowest speicifed delta or to a new_image.
+ *
+ * Parameters:
+ *  dev		ploop device (for online merge) or NULL (for offline merge)
+ *  start_level	start (lower) delta level to merge (online only)
+ *  end_level	end (higher) delta level to merge (online only)
+ *  raw		set to 1 if merging to base delta which is raw delta
+ *  merge_top	set to 1 is top rw delta is to be merged (online only)
+ *  images	list of delta files to be merged, from top to bottom
+ *  new_image	new image file name to merge to (NULL to merge to lowest delta)
+ *
+ * Returns: 0 or SYSEXIT_* error
+ */
 int merge_image(const char *device, int start_level, int end_level, int raw, int merge_top,
-		      char **images)
+		      char **images, const char *new_image)
 {
 	int last_delta = 0;
 	char **names = NULL;
@@ -302,13 +313,21 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 	__u32 blocksize = 0;
 	__u32 prev_blocksize = 0;
 	int version = PLOOP_FMT_UNDEFINED;
+	const char *merged_image;
 
-	if (start_level >= end_level || start_level < 0) {
-		ploop_err(0, "Invalid parameters: start_level %d end_level %d",
-			start_level, end_level);
-		return -1;
+	if (new_image && access(new_image, F_OK) == 0) {
+		ploop_err(EEXIST, "Can't merge to new image %s", new_image);
+		return SYSEXIT_PARAM;
 	}
+
 	if (device) {
+		/* Online merge */
+		if (start_level >= end_level || start_level < 0) {
+			ploop_err(0, "Invalid parameters: start_level %d end_level %d",
+					start_level, end_level);
+			return SYSEXIT_PARAM;
+		}
+
 		ret = ploop_complete_running_operation(device);
 		if (ret)
 			return ret;
@@ -323,22 +342,41 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 		} else
 			names = images;
 
-		if ((ret = grow_lower_delta(device, merge_top, start_level, end_level)))
-			return ret;
+		if (!new_image)
+			if ((ret = grow_lower_delta(device, merge_top,
+						start_level, end_level)))
+				return ret;
 
 		if (end_level == 0) {
 			int lfd;
 
+			if (new_image) {
+				/* Special case: only one delta below top one:
+				 * copy it to new_image and do ploop_replace()
+				 */
+				ret = copy_delta(names[0], new_image);
+				if (ret)
+					return ret;
+
+				ret = replace_delta(device, start_level, new_image);
+				if (ret)
+					goto rm_delta;
+			}
 			ploop_log(0, "Merging top delta");
 			lfd = open(device, O_RDONLY);
 			if (lfd < 0) {
 				ploop_err(errno, "open dev %s", device);
-				return SYSEXIT_DEVICE;
+				ret = SYSEXIT_DEVICE;
+				goto rm_delta;
 			}
 
 			ret = ioctl_device(lfd, PLOOP_IOC_MERGE, 0);
 			close(lfd);
 
+rm_delta:
+			if (ret && new_image)
+				unlink(new_image);
+
 			return ret;
 		}
 		last_delta = end_level - start_level;
@@ -347,6 +385,14 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 		names = images;
 	}
 
+	if (new_image) {
+		last_delta++;
+		merged_image = new_image;
+	}
+	else {
+		merged_image = names[last_delta];
+	}
+
 	init_delta_array(&da);
 
 	for (i = 0; i < last_delta; i++) {
@@ -361,6 +407,7 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 		if (i != 0 && blocksize != prev_blocksize) {
 			ploop_err(errno, "Wrong blocksize %s bs=%d [prev bs=%d]",
 					names[i], blocksize, prev_blocksize);
+			ret = SYSEXIT_PLOOPFMT;
 			goto merge_done2;
 		}
 		prev_blocksize = blocksize;
@@ -368,18 +415,34 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 		if (i != 0 && version != da.delta_arr[i].version) {
 			ploop_err(errno, "Wrong version %s %d [prev %d]",
 					names[i], da.delta_arr[i].version, version);
+			ret = SYSEXIT_PLOOPFMT;
 			goto merge_done2;
 		}
 		version = da.delta_arr[i].version;
 	}
 	if (blocksize == 0) {
 		ploop_err(errno, "Wrong blocksize 0");
+		ret = SYSEXIT_PLOOPFMT;
 		goto merge_done2;
 
 	}
 	cluster = S2B(blocksize);
+
+	if (new_image) { /* Create it */
+		struct ploop_pvd_header *vh;
+		off_t size;
+		int mode = (raw) ? PLOOP_RAW_MODE : PLOOP_EXPANDED_MODE;
+
+		vh = (struct ploop_pvd_header *)da.delta_arr[0].hdr0;
+		size = get_SizeInSectors(vh);
+
+		ret = create_image(new_image, blocksize, size, mode, version);
+		if (ret)
+			goto merge_done2;
+	}
+
 	if (!raw) {
-		if (open_delta(&odelta, names[last_delta], O_RDWR,
+		if (open_delta(&odelta, merged_image, O_RDWR,
 			       device ? OD_NOFLAGS : OD_OFFLINE)) {
 			ploop_err(errno, "open_delta");
 			ret = SYSEXIT_OPEN;
@@ -391,7 +454,7 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 			goto merge_done2;
 		}
 	} else {
-		if (open_delta_simple(&odelta, names[last_delta], O_RDWR,
+		if (open_delta_simple(&odelta, merged_image, O_RDWR,
 				      device ? 0 : OD_OFFLINE)) {
 			ret = SYSEXIT_WRITE;
 			goto merge_done2;
@@ -402,7 +465,7 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 		goto merge_done2;
 	}
 
-	if (!device) {
+	if (!device && !new_image) {
 		struct ploop_pvd_header *vh;
 		vh = (struct ploop_pvd_header *)da.delta_arr[0].hdr0;
 
@@ -414,12 +477,12 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 			off_t src_size = get_SizeInSectors(vh);
 			off_t dst_size;
 
-			ret = read_size_from_image(names[last_delta], 1, &dst_size);
+			ret = read_size_from_image(merged_image, 1, &dst_size);
 			if (ret)
 				goto merge_done;
 
 			if (src_size > dst_size) {
-				ret = grow_raw_delta(names[last_delta],
+				ret = grow_raw_delta(merged_image,
 					       S2B(src_size - dst_size));
 				if (ret)
 					goto merge_done;
@@ -463,10 +526,9 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 			 * lookup lower deltas.
 			 */
 			if (da.delta_arr[0].l2[k] == 0) {
-				if (locate_l2_entry(&da, 0, i, k, &level2)) {
-					ret = -1;
+				ret = locate_l2_entry(&da, 0, i, k, &level2);
+				if (ret)
 					goto merge_done;
-				}
 				if (level2 < 0)
 					continue;
 			}
@@ -508,7 +570,7 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 							blocksize, version);
 				if (odelta.l2[k] == 0) {
 					ploop_err(0, "abort: odelta.l2[k] == 0");
-					ret = -1;
+					ret = SYSEXIT_ABORT;
 					goto merge_done;
 				}
 				odelta.l2_dirty = 1;
@@ -523,14 +585,7 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 		}
 	}
 
-	if (device && allocated &&
-	    odelta.fops->update_size &&
-	    (ret = odelta.fops->update_size(odelta.fd, names[last_delta]))) {
-		ploop_err(errno, "update_size");
-		goto merge_done;
-	}
-
-	if (odelta.fops->fsync(odelta.fd)) {
+	if (fsync(odelta.fd)) {
 		ploop_err(errno, "fsync");
 		ret = SYSEXIT_FSYNC;
 		goto merge_done;
@@ -541,12 +596,12 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 
 		if (odelta.l2_cache < 0) {
 			ploop_err(0, "abort: odelta.l2_cache < 0");
-			ret = -1;
+			ret = SYSEXIT_ABORT;
 			goto merge_done;
 		}
 		if (odelta.l2_cache >= odelta.l1_size) {
 			ploop_err(0, "abort: odelta.l2_cache >= odelta.l1_size");
-			ret = -1;
+			ret = SYSEXIT_ABORT;
 			goto merge_done;
 		}
 
@@ -559,7 +614,7 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 			ret = SYSEXIT_WRITE;
 			goto merge_done;
 		}
-		if (odelta.fops->fsync(odelta.fd)) {
+		if (fsync(odelta.fd)) {
 			ploop_err(errno, "fsync");
 			ret = SYSEXIT_FSYNC;
 			goto merge_done;
@@ -570,7 +625,7 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 		ploop_err(errno, "clear_delta");
 		ret = SYSEXIT_WRITE;
 	} else {
-		if (odelta.fops->fsync(odelta.fd)) {
+		if (fsync(odelta.fd)) {
 			ploop_err(errno, "fsync");
 			ret = SYSEXIT_FSYNC;
 		}
@@ -590,6 +645,21 @@ merge_done:
 			goto merge_done2;
 		}
 
+		if (new_image) {
+			int fd;
+
+			fd = open(new_image, O_DIRECT | O_RDONLY);
+			if (fd < 0) {
+				ploop_err(errno, "Can't open %s", new_image);
+				ret = SYSEXIT_OPEN;
+				goto close_lfd;
+			}
+			ret = do_replace_delta(lfd, start_level, fd, blocksize, new_image);
+			close(fd);
+			if (ret)
+				goto close_lfd;
+		}
+
 		level = start_level + 1;
 
 		for (i = start_level + 1; i <= end_level; i++) {
@@ -604,7 +674,7 @@ merge_done:
 			ploop_log(0, "Merging top delta");
 			ret = ioctl_device(lfd, PLOOP_IOC_MERGE, 0);
 		}
-
+close_lfd:
 		close(lfd);
 	}
 merge_done2:
@@ -613,74 +683,113 @@ merge_done2:
 	return ret;
 }
 
-static const char *get_devs_str(char **devs, char *buf, int size)
+/* Only call after ploop_di_can_merge() returned 0 */
+int check_snapshot_mounts(struct ploop_disk_images_data *di,
+		const char *ancestor_guid, const char *descendant_guid)
 {
-	char **p;
-	char *sp = buf;
-	char *ep = buf + size;
+	const char *guid = descendant_guid;
 
-	for (p = devs; *p != NULL; p++) {
-		sp += snprintf(sp, ep - sp, "%s ", *p);
-		if (sp >= ep)
+	while (1) {
+		int idx, temp, ret;
+		const char *fname;
+
+		idx = find_snapshot_by_guid(di, guid);
+		if (idx == -1) {
+			ploop_err(0, "Can't find snapshot by guid %s",
+					guid);
+			return SYSEXIT_DISKDESCR;
+		}
+
+		fname = find_image_by_guid(di, guid);
+		if (!fname) {
+			ploop_err(0, "Can't find image by guid %s",
+					guid);
+			return SYSEXIT_DISKDESCR;
+		}
+
+		temp = di->snapshots[idx]->temporary;
+
+		ret = check_snapshot_mount(di, guid, fname, temp);
+		if (ret)
+			return ret;
+
+		if (!guidcmp(guid, ancestor_guid))
 			break;
-	}
-	return buf;
-}
 
-int check_snapshot_mount(struct ploop_disk_images_data *di,
-		int temporary, const char *parent_fname,
-		const char *child_fname, const char *child_guid)
-{
-	int ret = 0;
-	char **devs, **p;
-	char buf[512];
-
-	/* Check if upper (child) delta that will be merged and
-	 * destroyed is mounted
-	 */
-	if (guidcmp(child_guid, di->top_guid) != 0 &&
-			ploop_get_dev_by_delta(di->images[0]->file,
-				child_fname, NULL, &devs) == 0)
-	{
-		ploop_err(0, "Snapshot is busy by device(s): %s",
-				get_devs_str(devs, buf, sizeof(buf)));
-		ret = SYSEXIT_EBUSY;
-		goto out;
-	}
-
-	/* check if snapshot (parent) delta is mounted */
-	if (ploop_get_dev_by_delta(di->images[0]->file,
-			parent_fname, NULL, &devs) == 0)
-	{
-		/* unmount temporary snapshot */
-		if (temporary) {
-			for (p = devs; *p != NULL; p++) {
-				ret = ploop_umount(*p, NULL);
-				if (ret)
-					break;
-			}
-		} else {
-			ploop_err(0, "Snapshot is busy by device(s): %s",
-					get_devs_str(devs, buf, sizeof(buf)));
-			ret = SYSEXIT_EBUSY;
+		guid = di->snapshots[idx]->parent_guid;
+		if (!strcmp(guid, NONE_UUID)) {
+			/* can't happen */
+			ploop_err(0, "Unexpectedly found %s to be base", guid);
+			return SYSEXIT_DISKDESCR;
 		}
 	}
-out:
-	ploop_free_array(devs);
-	return ret;
+
+	return 0;
 }
 
-int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid, int merge_mode)
+/* Logs a line showing deltas to merge and the merge destination,
+ * that looks something like these:
+ *
+ *	Merging image: delta_file -> parent_delta_file
+ *
+ *	Merging images: delta_file_1 file_2 file_3 -> (new) new_delta
+ */
+static void log_merge_images_info(struct ploop_disk_images_data *di,
+		char **names, const char *new_delta)
+{
+	char basedir[PATH_MAX];
+	char imglist[LOG_BUF_SIZE];
+	char merged_image[PATH_MAX];
+	int i, pos = 0, nimg;
+
+	get_basedir(di->runtime->xml_fname, basedir, sizeof(basedir));
+
+	nimg = get_list_size(names) - 1;
+	if (new_delta) {
+		nimg++;
+		// merged_image = new_delta;
+		normalize_image_name(basedir, new_delta,
+				merged_image, sizeof(merged_image));
+	}
+	else {
+		// merged_image = names[nimg];
+		normalize_image_name(basedir, names[nimg],
+				merged_image, sizeof(merged_image));
+	}
+
+	for (i = 0; i < nimg; i++) {
+		int n;
+		char img[PATH_MAX];
+
+		normalize_image_name(basedir, names[i], img, sizeof(img));
+		n = snprintf(imglist + pos, sizeof(imglist) - pos, "%s ", img);
+		if (n <= 0)
+			// error?
+			break;
+		pos += n;
+		if (pos >= sizeof(imglist))
+			// output truncated!
+			break;
+	}
+
+	ploop_log(0, "Merging image%s: %s-> %s%s",
+			nimg > 1 ? "s" : "",
+			imglist,
+			new_delta ? "(new) " : "", merged_image);
+}
+
+int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di,
+		const char *guid, int merge_mode, const char *new_delta)
 {
 	char conf[PATH_MAX];
+	char conf_tmp[PATH_MAX];
 	char dev[64];
 	char *device = NULL;
-	char *fname = NULL;
 	char *parent_fname = NULL;
 	char *child_fname = NULL;
 	char *delete_fname = NULL;
-	char *child_guid = NULL;
-	char *parent_guid = NULL;
+	const char *child_guid = NULL;
+	const char *parent_guid = NULL;
 	char *names[3] = {};
 	int ret;
 	int start_level = -1;
@@ -688,78 +797,69 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 	int merge_top = 0;
 	int raw = 0;
 	int online = 0;
-	int snap_idx;
-	int temporary = 0;
+	int parent_idx, child_idx; /* parent and child snapshot ids */
 	struct merge_info info = {};
 	int i, nelem;
 
 	ret = SYSEXIT_PARAM;
-	snap_idx = find_snapshot_by_guid(di, guid);
-	if (snap_idx == -1) {
-		ploop_err(0, "Can't find snapshot by uuid %s",
-				guid);
-		return ret;
-	}
 
-	temporary = di->snapshots[snap_idx]->temporary;
-
-	fname = find_image_by_guid(di, guid);
-	if (fname == NULL) {
-		ploop_err(0, "Can't find image by uuid %s",
-				guid);
-		return ret;
-	}
 	if (merge_mode == PLOOP_MERGE_WITH_CHILD) {
-		parent_guid = (char *)guid;
-		parent_fname = fname;
-		ret = ploop_get_child_by_uuid(di, guid, &child_guid);
-		if (ret) {
-			ploop_err(0, "Can't find child by uuid %s",
-					child_guid);
+		parent_guid = guid;
+		child_guid = ploop_find_child_by_guid(di, guid);
+		if (!child_guid) {
+			ploop_err(0, "Can't find child of uuid %s", guid);
 			goto err;
 		}
-		child_fname = find_image_by_guid(di, child_guid);
-		if (child_fname == NULL) {
-			ploop_err(0, "Can't find image by uuid %s",
-					child_guid);
-			goto err;
-		}
-		temporary = di->snapshots[snap_idx]->temporary;
 	} else if (merge_mode == PLOOP_MERGE_WITH_PARENT) {
-		parent_guid = di->snapshots[snap_idx]->parent_guid;
-		child_guid = (char *)guid;
-		if (strcmp(parent_guid, NONE_UUID) == 0) {
-			ploop_err(0, "Unable to merge base image");
+		child_guid = guid;
+		parent_guid = ploop_find_parent_by_guid(di, guid);
+		if (!parent_guid) {
+			ploop_err(0, "Can't find parent of uuid %s "
+					"(is that a base image?)", guid);
 			goto err;
 		}
-		child_fname = fname;
-		parent_fname = find_image_by_guid(di, parent_guid);
-		if (parent_fname == NULL) {
-			ploop_err(0, "Can't find image by uuid %s",
-					 parent_guid);
-			goto err;
-		}
-		snap_idx = find_snapshot_by_guid(di, parent_guid);
-		if (snap_idx == -1) {
-			ploop_err(0, "Can't find snapshot by uuid %s",
-					parent_guid);
-
-			goto err;
-		}
-		temporary = di->snapshots[snap_idx]->temporary;
 	} else
 		assert(0);
 
+	child_fname = find_image_by_guid(di, child_guid);
+	if (child_fname == NULL) {
+		ploop_err(0, "Can't find image by uuid %s",
+				child_guid);
+		goto err;
+	}
+	parent_fname = find_image_by_guid(di, parent_guid);
+	if (parent_fname == NULL) {
+		ploop_err(0, "Can't find image by uuid %s",
+				parent_guid);
+		goto err;
+	}
+	parent_idx = find_snapshot_by_guid(di, parent_guid);
+	if (parent_idx == -1) {
+		ploop_err(0, "Can't find snapshot by uuid %s",
+				parent_guid);
+
+		goto err;
+	}
+	child_idx = find_snapshot_by_guid(di, child_guid);
+	if (child_idx == -1) {
+		ploop_err(0, "Can't find snapshot by uuid %s",
+				child_guid);
+
+		goto err;
+	}
+
 	nelem = ploop_get_child_count_by_uuid(di, parent_guid);
 	if (nelem > 1) {
-		ploop_err(0, "Unable to merge snapshot %s: it has %d children",
+		ploop_err(0, "Can't merge to snapshot %s: it has %d children",
 				parent_guid, nelem);
 		return SYSEXIT_PARAM;
 	}
 
 	ret = ploop_find_dev_by_dd(di, dev, sizeof(dev));
-	if (ret == -1)
+	if (ret == -1) {
+		ret = SYSEXIT_SYS;
 		goto err;
+	}
 	else if (ret == 0)
 		online = 1;
 
@@ -787,7 +887,7 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 
 		if (end_level != -1 && start_level != -1) {
 			if (end_level != start_level + 1) {
-				ploop_err(0, "Inconsistensy detected %s [%d] %s [%d]",
+				ploop_err(0, "Inconsistency detected %s [%d] %s [%d]",
 						parent_fname, end_level, child_fname, start_level);
 				ret = SYSEXIT_PARAM;
 				goto err;
@@ -799,7 +899,7 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 		} else if (end_level == -1 && start_level == -1) {
 			online = 0;
 		} else {
-			ploop_err(0, "Inconsistensy detected %s [%d] %s [%d]",
+			ploop_err(0, "Inconsistency detected %s [%d] %s [%d]",
 					parent_fname, end_level, child_fname, start_level);
 			ret = SYSEXIT_PARAM;
 			goto err;
@@ -811,11 +911,11 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 		end_level = 1;
 		/* Only base image could be in RAW format */
 		if (di->mode == PLOOP_RAW_MODE &&
-				!guidcmp(di->snapshots[snap_idx]->parent_guid, NONE_UUID))
+				!guidcmp(di->snapshots[parent_idx]->parent_guid, NONE_UUID))
 			raw = 1;
 	}
 
-	names[0] = strdup(child_fname);
+	names[0] = delete_fname = strdup(child_fname);
 	if (names[0] == NULL) {
 		ret = SYSEXIT_MALLOC;
 		goto err;
@@ -827,46 +927,396 @@ int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *
 	}
 	names[2] = NULL;
 
-	ret = check_snapshot_mount(di, temporary, parent_fname,
-			child_fname, child_guid);
+	ret = check_snapshot_mount(di, parent_guid, parent_fname,
+			di->snapshots[parent_idx]->temporary);
+	if (ret)
+		goto err;
+	ret = check_snapshot_mount(di, child_guid, child_fname,
+			di->snapshots[child_idx]->temporary);
 	if (ret)
 		goto err;
 
-	ploop_log(0, "%sline merge uuid %s -> %s image %s -> %s %s",
+	ploop_log(0, "%sline merge %s -> %s%s",
 			online ? "On": "Off",
 			child_guid, parent_guid,
-			names[0], names[1],
-			raw ? "raw" : "");
+			raw ? " (raw)" : "");
+
+	log_merge_images_info(di, names, new_delta);
 
 	/* make validation before real merge */
-	ret = ploop_di_merge_image(di, child_guid, &delete_fname);
-	if (ret)
-		goto err;
-
-
-	ret = merge_image(device, start_level, end_level, raw, merge_top, names);
+	ret = ploop_di_merge_image(di, child_guid);
 	if (ret)
 		goto err;
 
 	get_disk_descriptor_fname(di, conf, sizeof(conf));
-	ret = ploop_store_diskdescriptor(conf, di);
+	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
+	ret = ploop_store_diskdescriptor(conf_tmp, di);
 	if (ret)
 		goto err;
+
+	ret = merge_image(device, start_level, end_level, raw, merge_top, names, new_delta);
+	if (ret)
+		goto err;
+
+	if (new_delta) {
+		/* Write new delta name to dd.xml, and remove the old file.
+		 * Note we can only write new delta now after merge_image()
+		 * as the file is created and we can use realpath() on it.
+		 */
+		int idx;
+		char *oldimg, *newimg;
+
+		newimg = realpath(new_delta, NULL);
+		if (!newimg) {
+			ploop_err(errno, "Error in realpath(%s)", new_delta);
+			ret = SYSEXIT_PARAM;
+			goto err;
+		}
+
+		idx = find_image_idx_by_guid(di, child_guid);
+		if (idx == -1) {
+			ploop_err(0, "Unable to find image by uuid %s",
+					child_guid);
+			ret = SYSEXIT_PARAM;
+			goto err;
+		}
+
+		oldimg = di->images[idx]->file;
+		di->images[idx]->file = newimg;
+
+		ret = ploop_store_diskdescriptor(conf_tmp, di);
+		if (ret)
+			goto err;
+
+		ploop_log(0, "Removing %s", oldimg);
+		ret = unlink(oldimg);
+		if (ret) {
+			ploop_err(errno, "unlink %s", oldimg);
+			ret = SYSEXIT_UNLINK;
+		}
+
+		free(oldimg);
+	}
+
+	if (rename(conf_tmp, conf)) {
+		ploop_err(errno, "Can't rename %s %s",
+				conf_tmp, conf);
+		ret = SYSEXIT_RENAME;
+		goto err;
+	}
+
 	ploop_log(0, "Removing %s", delete_fname);
 	if (unlink(delete_fname)) {
 		ploop_err(errno, "unlink %s", delete_fname);
 		ret = SYSEXIT_UNLINK;
 	}
+
 	if (ret == 0)
 		ploop_log(0, "ploop %s %s has been successfully merged",
-				get_snap_str(temporary), guid);
+				get_snap_str(di->snapshots[parent_idx]->temporary),
+				parent_guid);
 
 err:
 	for (i = 0; names[i] != NULL; i++)
 		free(names[i]);
 
-	free(delete_fname);
-	free_images_list(info.names);
+	ploop_free_array(info.names);
+
+	return ret;
+}
+
+/* Given a device and any two deltas, figure out if a merge between these
+ * is online or offline. For online case, fill in the merge parameters.
+ *
+ * Parameters:
+ *   dev	ploop device
+ *   anc_fname	ancestor (grand parent) image file name
+ *   dsc_fname	descentant (grand child) image file name
+ *
+ * Returns:
+ *   0 or SYSEXIT_* error
+ *
+ * Returned data:
+ *   *online		1 if this is online merge and the following data:
+ *   mi->start_level	start level to merge
+ *   mi->end_level	end level to merge
+ *   mi->raw		whether base delta is raw
+ *   mi->top_level	device top level
+ *   mi->merge_top	whether top level is affected
+ *   mi->names		list of deltas to merge, from descendant to ancestor
+ */
+static int prepare_online_merge(const char *dev,
+		const char *anc_fname, const char *dsc_fname,
+		int *online, struct merge_info *mi)
+{
+	struct merge_info info = {};
+	int i;
+	int anc_idx = -1, dsc_idx = -1; /* indexes in info */
+	int ret;
+
+	ret = get_delta_info(dev, &info);
+	if (ret)
+		return ret;
+
+	ret = SYSEXIT_FSTAT;
+	/* Try to find both images among deltas in running ploop */
+	for (i = 0; info.names[i] != NULL; i++) {
+		int r;
+
+		if (dsc_idx == -1) {
+			r = ploop_fname_cmp(info.names[i], dsc_fname);
+			if (r == -1)
+				goto out;
+			else if (r == 0) {
+				dsc_idx = i;
+				continue;
+			}
+		}
+		if (anc_idx == -1) {
+			r = ploop_fname_cmp(info.names[i], anc_fname);
+			if (r == -1)
+				goto out;
+			else if (r == 0) {
+				anc_idx = i;
+				continue;
+			}
+		}
+	}
+
+	/* Figure out if it's online, offline, or mixed case */
+	if (anc_idx != -1 && dsc_idx != -1) {
+		/* found both deltas: ONLINE */
+		int nelem, i, idx;
+
+		*online = 1;
+		nelem = get_list_size(info.names);
+		mi->start_level = nelem - anc_idx - 1;
+		mi->end_level = nelem - dsc_idx - 1;
+		if (mi->end_level <= mi->start_level) {
+			ploop_err(0, "Inconsistency: end_level <= start_level"
+					" (%d <= %d)",
+					mi->end_level, mi->start_level);
+			ret = SYSEXIT_PARAM;
+			goto out;
+		}
+
+		mi->names = calloc(mi->end_level - mi->start_level + 2,
+				sizeof(char *));
+		if (!mi->names) {
+			ploop_err(errno, "Can't malloc");
+			ret = SYSEXIT_MALLOC;
+			goto out;
+		}
+		/* Copy part of names we are interested in */
+		for (i = 0, idx = dsc_idx; idx <= anc_idx; idx++, i++)
+			mi->names[i] = strdup(info.names[idx]);
+
+		/* set the needed flags */
+		mi->raw = (mi->start_level == 0) ? info.raw : 0;
+		mi->top_level = info.top_level;
+		mi->merge_top = (info.top_level == mi->end_level);
+
+		ret = 0;
+	}
+	else if (anc_idx == -1 && dsc_idx == -1) {
+		/* no used deltas found: OFFLINE */
+		*online = 0;
+		ret = 0;
+	}
+	else {
+		/* Some deltas found: mixed case */
+		ploop_err(0, "Can't do mixed merge (some deltas "
+				"are online, some are offline)");
+		ret = SYSEXIT_PARAM;
+	}
+out:
+	ploop_free_array(info.names);
+
+	return ret;
+}
+
+/* Given two snapshot guids, fill in the merge info (for offline case)
+ *
+ * Parameters:
+ *   di			disk descriptor info
+ *   ancestor_guid	ancestor (grand parent, lower) guid
+ *   descendant_guid	descentant (grand child, higher) guid
+ *
+ * Returns:
+ *   0 or SYSEXIT_* error
+ *
+ * Returned data:
+ *   mi->raw		whether base delta is raw
+ *   mi->names		list of deltas to merge, from descentant to ancestor
+ */
+static int prepare_offline_merge(struct ploop_disk_images_data *di,
+		const char *ancestor_guid, const char *descendant_guid,
+		struct merge_info *mi)
+{
+	int ret = 0;
+
+	mi->names = make_images_list_by_guids(di,
+			descendant_guid, ancestor_guid, 1);
+	if (!mi->names)
+		/* error is printed by make_images_list_by_guids */
+		return SYSEXIT_SYS;
+
+	if (di->mode == PLOOP_RAW_MODE) {
+		int i;
+
+		/* We should set raw=1 if merging down to base image */
+		i = find_snapshot_by_guid(di, ancestor_guid);
+		if (i == -1) { /* should not ever happen */
+			ploop_err(0, "Can't find snapshot by guid %s",
+					ancestor_guid);
+			ret = SYSEXIT_PARAM;
+			goto out;
+		}
+		if (strcmp(di->snapshots[i]->parent_guid, NONE_UUID) == 0)
+			mi->raw = 1;
+	}
+
+out:
+	if (ret && mi->names)
+		ploop_free_array(mi->names);
+
+	return ret;
+}
+
+/* Merge multiple snapshots by a range of GUIDs */
+int ploop_merge_snapshots_by_guids(struct ploop_disk_images_data *di,
+		const char *descendant_guid, const char *ancestor_guid,
+		const char *new_delta)
+{
+	int i, ret;
+	int online;
+	char dev[64] = "";
+	struct merge_info info = {};
+	char conf[PATH_MAX];
+	char conf_tmp[PATH_MAX];
+
+	/* Can we merge these? */
+	ret = ploop_di_can_merge_images(di, ancestor_guid, descendant_guid);
+	if (ret)
+		return ret;
+
+	/* Is it online or offline merge? */
+	ret = ploop_find_dev_by_dd(di, dev, sizeof(dev));
+	if (ret == -1)
+		return SYSEXIT_SYS;
+	online = !ret;
+	if (online) {
+		const char *anc_fname, *dsc_fname;
+
+		anc_fname = find_image_by_guid(di, ancestor_guid);
+		dsc_fname = find_image_by_guid(di, descendant_guid);
+		/* Can be online, offline, or mixed */
+		ret = prepare_online_merge(dev, anc_fname, dsc_fname,
+				&online, &info);
+		if (ret)
+			return ret;
+	}
+
+	if (!online) {
+		ret = prepare_offline_merge(di,
+				ancestor_guid, descendant_guid, &info);
+		if (ret)
+			goto out;
+	}
+
+	/* FIXME: check if any snapshots are mounted */
+	ret = check_snapshot_mounts(di,
+			ancestor_guid, descendant_guid);
+	if (ret)
+		goto out;
+
+	i = get_list_size(info.names);
+	ploop_log(0, "%s merge of %d snapshot%s",
+			online ? "Online" : "Offline",
+			i - 1, i < 3 ? "" : "s");
+	log_merge_images_info(di, info.names, new_delta);
+
+	/* First, merge in dd.xml (and do some extra checks */
+	ret = ploop_di_merge_images(di, descendant_guid, i - 1);
+	if (ret)
+		goto out;
+
+	get_disk_descriptor_fname(di, conf, sizeof(conf));
+	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
+	ret = ploop_store_diskdescriptor(conf_tmp, di);
+	if (ret)
+		goto out;
+
+	/* Do the actual merge */
+	ret = merge_image(online ? dev : NULL,
+			info.start_level, info.end_level,
+			info.raw, info.merge_top, info.names, new_delta);
+	if (ret)
+		goto out;
+
+	if (new_delta) {
+		/* Write new delta name to dd.xml, and remove the old file.
+		 * Note we can only write new delta now after merge_image()
+		 * as the file is created and we can use realpath() on it.
+		 */
+		int idx;
+		char *oldimg, *newimg;
+
+		newimg = realpath(new_delta, NULL);
+		if (!newimg) {
+			ploop_err(errno, "Error in realpath(%s)", new_delta);
+			ret = SYSEXIT_PARAM;
+			goto out;
+		}
+
+		idx = find_image_idx_by_guid(di, descendant_guid);
+		if (idx == -1) {
+			ploop_err(0, "Can't find image by uuid %s",
+					descendant_guid);
+			ret = SYSEXIT_PARAM;
+			goto out;
+		}
+
+		oldimg = di->images[idx]->file;
+		di->images[idx]->file = newimg;
+
+		ret = ploop_store_diskdescriptor(conf_tmp, di);
+		if (ret)
+			goto out;
+
+		ploop_log(0, "Removing %s", oldimg);
+		ret = unlink(oldimg);
+		if (ret) {
+			ploop_err(errno, "Can't remove %s", oldimg);
+			ret = SYSEXIT_UNLINK;
+		}
+
+		free(oldimg);
+	}
+
+	/* Put a new dd.xml */
+	if (rename(conf_tmp, conf)) {
+		ploop_err(errno, "Can't rename %s %s", conf_tmp, conf);
+		ret = SYSEXIT_RENAME;
+		goto out;
+	}
+
+	/* Remove images which were merged, i.e. all but the lowest one */
+	for (i = 0; info.names[i+1]; i++) {
+		const char *file = info.names[i];
+
+		ploop_log(0, "Removing %s", file);
+		if (unlink(file)) {
+			ploop_err(errno, "Can't remove %s", file);
+			ret = SYSEXIT_UNLINK;
+		}
+	}
+
+	if (ret == 0)
+		ploop_log(0, "Snapshots successfully merged");
+
+out:
+	ploop_free_array(info.names);
 
 	return ret;
 }
@@ -879,20 +1329,24 @@ int ploop_merge_snapshot(struct ploop_disk_images_data *di, struct ploop_merge_p
 	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
-	if (param->guid != NULL)
+	if (param->guid2 != NULL) {
+		/* merge multiple snapshots at once */
+		ret = ploop_merge_snapshots_by_guids(di, param->guid, param->guid2, param->new_delta);
+		goto unlock;
+	}
+	else if (param->guid != NULL)
 		guid = param->guid;
 	else if (!param->merge_all)
 		guid = di->top_guid;
 
 	if (guid != NULL) {
-		ret = ploop_merge_snapshot_by_guid(di, guid, PLOOP_MERGE_WITH_PARENT);
+		ret = ploop_merge_snapshot_by_guid(di, guid, PLOOP_MERGE_WITH_PARENT, param->new_delta);
 	} else {
-		while (di->nsnapshots != 1) {
-			ret = ploop_merge_snapshot_by_guid(di, di->top_guid, PLOOP_MERGE_WITH_PARENT);
-			if (ret)
-				break;
-		}
+		/* merge all from top to base */
+		guid = get_base_delta_uuid(di);
+		ret =  ploop_merge_snapshots_by_guids(di, di->top_guid, guid, param->new_delta);
 	}
+unlock:
 	ploop_unlock_dd(di);
 
 	return ret;
