@@ -97,7 +97,20 @@ int get_list_size(char **list)
 	return i;
 }
 
-char **make_images_list(struct ploop_disk_images_data *di, const char *guid, int reverse)
+/* Given di and two guids, returns a list of images (delta files)
+ * corresponding to these guids.
+ *
+ * Parameters:
+ *  di		dd.xml data
+ *  guid	guid to start from (top one)
+ *  end_guid	guid to stop at, or NULL to stop at base
+ *  reverse	1 to list from base to top, 0 from top
+ *
+ * Returns a pointer to array of char*, having NULL as last element.
+ * Note ploop_free_array() should be used to free memory.
+ */
+char **make_images_list_by_guids(struct ploop_disk_images_data *di,
+		const char *guid, const char *end_guid, int reverse)
 {
 	int n;
 	char **images;
@@ -134,14 +147,22 @@ char **make_images_list(struct ploop_disk_images_data *di, const char *guid, int
 			ploop_err(0, "Inconsistency detected: snapshots > images");
 			goto err;
 		}
+		if (done)
+			break;
 		guid = di->snapshots[snap_id]->parent_guid;
-		if (!strcmp(guid, NONE_UUID)) {
+		if (end_guid && !strcmp(guid, end_guid))
+			done = 1; /* still need to add the last image */
+		else if (!strcmp(guid, NONE_UUID)) {
 			done = 1;
 			break;
 		}
 	}
 	if (!done) {
-		ploop_err(0, "Inconsistency detected, base image not found");
+		if (end_guid)
+			ploop_err(0, "Snapshot %s not found", end_guid);
+		else
+			ploop_err(0, "Inconsistency detected: base image not found");
+
 		goto err;
 	}
 	images[++n] = NULL;
@@ -159,20 +180,13 @@ char **make_images_list(struct ploop_disk_images_data *di, const char *guid, int
 
 err:
 	images[n] = NULL;
-	free_images_list(images);
+	ploop_free_array(images);
 	return NULL;
 }
 
-void free_images_list(char **images)
+char **make_images_list(struct ploop_disk_images_data *di, const char *guid, int reverse)
 {
-	int i;
-
-	if (images == NULL)
-		return;
-	for (i = 0; images[i] != NULL; i++)
-		free(images[i]);
-
-	free(images);
+	return make_images_list_by_guids(di, guid, NULL, reverse);
 }
 
 static int WRITE(int fd, void * buf, unsigned int size)
@@ -194,7 +208,7 @@ int PWRITE(struct delta * delta, void * buf, unsigned int size, off_t off)
 {
 	ssize_t res;
 
-	res = delta->fops->pwrite(delta->fd, buf, size, off);
+	res = pwrite(delta->fd, buf, size, off);
 	if (res == size)
 		return 0;
 	if (res >= 0)
@@ -208,7 +222,7 @@ int PREAD(struct delta * delta, void *buf, unsigned int size, off_t off)
 {
 	ssize_t res;
 
-	res = delta->fops->pread(delta->fd, buf, size, off);
+	res = pread(delta->fd, buf, size, off);
 	if (res == size)
 		return 0;
 	if (res >= 0)
@@ -575,7 +589,7 @@ static void fill_diskdescriptor(struct ploop_pvd_header *vh, struct ploop_disk_i
 	di->sectors = vh->m_Sectors;
 }
 
-static int create_image(const char *file, __u32 blocksize, off_t size_sec, int mode,
+int create_image(const char *file, __u32 blocksize, off_t size_sec, int mode,
 		int version)
 {
 	int fd = -1;
@@ -975,7 +989,7 @@ static int delete_deltas(int devfd, const char *devname)
 {
 	int top;
 
-	if (ploop_get_top_level(devfd, devname, &top))
+	if (ploop_get_attr(devname, "top", &top))
 		return errno;
 
 	while (top >= 0) {
@@ -1008,6 +1022,40 @@ static int ploop_stop(int fd, const char *devname)
 	return 0;
 }
 
+/* Convert escape sequences used in /proc/mounts, /etc/mtab
+ * and /proc/self/mountinfo files, such as:
+ *
+ *	\040 -> space
+ *	\011 -> tab
+ *	\012 -> newline
+ *	\134 -> \
+ *
+ * Taken as is from util-linux-2.24.2, licensed under GNU LGPL 2.1
+ */
+#define isoctal(a)		(((a) & ~7) == '0')
+void unmangle_to_buffer(const char *s, char *buf, size_t len)
+{
+	size_t sz = 0;
+
+	if (!s)
+		return;
+
+	while(*s && sz < len - 1) {
+		if (*s == '\\' && sz + 3 < len - 1 && isoctal(s[1]) &&
+				isoctal(s[2]) && isoctal(s[3])) {
+
+			*buf++ = 64*(s[1] & 7) + 8*(s[2] & 7) + (s[3] & 7);
+			s += 4;
+			sz += 4;
+		} else {
+			*buf++ = *s++;
+			sz++;
+		}
+	}
+	*buf = '\0';
+}
+#undef isoctal
+
 /* Returns:
  *  0 mount point is found and saved to *out
  *  1 mount point not found (fs not mounted)
@@ -1041,8 +1089,7 @@ static int get_mount_dir(const char *device, char *out, int size)
 		if (_major == PLOOP_DEV_MAJOR &&
 				(_minor == minor || _minor == minor + 1))
 		{
-			strncpy(out, target, size - 1);
-			out[size - 1] = 0;
+			unmangle_to_buffer(target, out, size);
 			ret = 0;
 			break;
 		}
@@ -1120,7 +1167,7 @@ int ploop_get_dev_by_mnt(const char *path, char *buf, int size)
 	return get_dev_by_mnt(path, 1, buf, size);
 }
 
-static char *get_base_delta_uuid(struct ploop_disk_images_data *di)
+char *get_base_delta_uuid(struct ploop_disk_images_data *di)
 {
 	int i;
 
@@ -1222,6 +1269,9 @@ static int ploop_mount_fs(struct ploop_mount_param *param)
 		if (e2fsck(part_device, E2FSCK_PREEN, &param->fsck_rc))
 			return SYSEXIT_FSCK;
 
+	if (param->target == NULL)
+		return 0;
+
 	/* Two step mount
 	 * 1 mount ro and read balloon inode
 	 * 2 remount with balloon_ino=ino
@@ -1258,9 +1308,8 @@ static int ploop_mount_fs(struct ploop_mount_param *param)
 static void print_sys_block_ploop(void)
 {
 	print_output(-1, "find",
-			"/sys/block/ploop[0-9]*/ -type f "
-			"-not -name '*event' "
-			"-not -name ve_device_add "
+			"/sys/block/ploop[0-9]*/pdelta/ -type f "
+			"\\( -name image -or -name io -or -name ro \\) "
 			"| xargs tail | grep -v '^$'");
 }
 
@@ -1295,13 +1344,36 @@ out:
 	return ret;
 }
 
+int do_replace_delta(int devfd, int level, int imgfd, __u32 blocksize,
+		const char *image)
+{
+	struct ploop_ctl_delta req = {};
+
+	req.c.pctl_cluster_log = ffs(blocksize) - 1;
+	req.c.pctl_level = level;
+	req.c.pctl_chunks = 1;
+	req.c.pctl_format = PLOOP_FMT_PLOOP1;
+	req.c.pctl_flags = PLOOP_FMT_RDONLY;
+
+	req.f.pctl_type = PLOOP_IO_AUTO;
+	req.f.pctl_fd = imgfd;
+
+	if (ioctl(devfd, PLOOP_IOC_REPLACE_DELTA, &req) < 0) {
+		ploop_err(errno, "Can't replace image %s", image);
+		if (errno == EBUSY)
+			print_sys_block_ploop();
+		return SYSEXIT_DEVIOC;
+	}
+
+	return 0;
+}
+
 int replace_delta(const char *device, int level, const char *image)
 {
 	int fd = -1, lfd = -1;
 	int top_level = 0;
 	int ret;
 	__u32 blocksize = 0;
-	struct ploop_ctl_delta req = {};
 
 	fd = open(image, O_DIRECT | O_RDONLY);
 	if (fd < 0) {
@@ -1332,23 +1404,7 @@ int replace_delta(const char *device, int level, const char *image)
 		goto out;
 	}
 
-	req.c.pctl_cluster_log = ffs(blocksize) - 1;
-	req.c.pctl_level = level;
-	req.c.pctl_chunks = 1;
-	req.c.pctl_format = PLOOP_FMT_PLOOP1;
-	req.c.pctl_flags = PLOOP_FMT_RDONLY;
-
-	req.f.pctl_type = PLOOP_IO_AUTO;
-	req.f.pctl_fd = fd;
-
-	if (ioctl(lfd, PLOOP_IOC_REPLACE_DELTA, &req) < 0) {
-		if (errno == EBUSY)
-			print_sys_block_ploop();
-		ploop_err(errno, "Can't replace image %s", image);
-		ret = SYSEXIT_DEVIOC;
-		goto out;
-	}
-	ret = 0;
+	ret = do_replace_delta(lfd, level, fd, blocksize, image);
 
 out:
 	if (lfd >= 0)
@@ -1359,20 +1415,65 @@ out:
 	return ret;
 }
 
+/* Check if f1 can be rename()d to f2.
+ * This is not a thorough check, currenty it just checks
+ * both files are on the same file system.
+ *
+ * Returns:
+ *  1	rename() will likely fail
+ *  0	rename() will likely succeed
+ * -1	internal error
+ */
+static int cant_rename(const char *f1, const char *f2) {
+	struct stat st1, st2;
+
+	if (lstat(f1, &st1)) {
+		ploop_err(errno, "Can't stat %s", f1);
+		return -1;
+	}
+	if (lstat(f2, &st2)) {
+		if (errno == ENOENT)
+			return 0; /* can rename */
+		ploop_err(errno, "Can't stat %s", f2);
+		return -1;
+	}
+	if (st1.st_dev != st2.st_dev) {
+		ploop_err(0, "Files %s and %s are on different file systems, "
+				"can't rename", f1, f2);
+		return 1; /* rename will return EXDEV */
+	}
+
+	/* FIXME: any other checks to add? */
+	return 0;
+}
+
+/* Return number of hardlinks for the file */
+static int st_nlink(const char *file)
+{
+	struct stat st;
+
+	if (stat(file, &st))
+		return -1;
+
+	return st.st_nlink;
+}
+
 int ploop_replace_image(struct ploop_disk_images_data *di,
 		struct ploop_replace_param *param)
 {
 	char dev[PATH_MAX];
-	char *file = NULL, *oldfile;
+	char *file = NULL, *oldfile, *tmp;
 	char conf[PATH_MAX], conf_tmp[PATH_MAX] = "";
-	int ret, level;
+	int ret, idx, level;
+	int keep_name = (param->flags & PLOOP_REPLACE_KEEP_NAME);
+	int offline = 0;
 
 	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
 	ret = SYSEXIT_PARAM;
 
-	if (!param || !param->file) {
+	if (!param->file) {
 		ploop_err(0, "New image file not specified");
 		goto err;
 	}
@@ -1384,83 +1485,192 @@ int ploop_replace_image(struct ploop_disk_images_data *di,
 	}
 
 	if (ploop_find_dev_by_dd(di, dev, sizeof(dev))) {
-		ploop_err(0, "Can't find running ploop device");
-		goto err;
+		ploop_log(1, "Can't find running ploop device, "
+				"doing offline replace");
+		offline = 1;
 	}
 
-	/* Image to be replaced is specified by either guid or level
+	/* Image to be replaced is specified by one of the following
+	 * (in the order of preference):
+	 *  1 guid
+	 *  2 current image file name
+	 *  3 level
+	 *
+	 * Try to find out level (if not set) and idx (as in di->images[idx])
 	 */
+	level = idx = -1;
 	if (param->guid) {
 		if (!is_valid_guid(param->guid)) {
 			ploop_err(0, "Invalid guid specified: %s", param->guid);
 			goto err;
 		}
 
-		level = find_image_idx_by_guid(di, param->guid);
-		if (level == -1) {
+		idx = find_image_idx_by_guid(di, param->guid);
+		if (idx == -1) {
 			ploop_err(0, "Can't find image by guid %s", param->guid);
 			goto err;
 		}
 	}
 	else if (param->cur_file) {
-		level = find_level_by_delta(dev, param->cur_file);
-		if (level < 0) {
-			ploop_err(0, "Can't find level by delta file name %s",
+		idx = find_image_idx_by_file(di, param->cur_file);
+		if (idx == -1) {
+			ploop_err(0, "Can't find image %s "
+					"in DiskDescriptor.xml",
 					param->cur_file);
 			goto err;
 		}
 	}
-	else {
+	else { /* by param->level */
+		char img[PATH_MAX];
+
+		if (offline) {
+			ploop_err(0, "Can't specify level for "
+					"offline replace");
+			goto err;
+		}
 		level = param->level;
+		/* Proper level check (against top_level) is to be done later
+		 * in replace_delta(). Here is just some basic sanity check.
+		 */
+		if (level < 0 || level >= di->nimages) {
+			ploop_err(0, "Invalid level %d", level);
+			goto err;
+		}
+
+		/* get delta file name and figure out idx */
+		if (ploop_get_delta_attr_str(dev, level, "image",
+					img, sizeof(img))) {
+			ret = SYSEXIT_SYSFS;
+			goto err;
+		}
+
+		idx = find_image_idx_by_file(di, img);
+		if (idx < 0) {
+			ploop_err(0, "Can't find image %s "
+					"in DiskDescriptor.xml", img);
+			/* This could only happen if dd.xml is wrong/bad */
+			ret = SYSEXIT_DISKDESCR;
+			goto err;
+		}
 	}
-	/* Proper level check (against top_level) is to be done later
-	 * in replace_delta(). Here is just some basic sanity check.
-	 * */
-	if (level < 0 || level >= di->nimages) {
-		ploop_err(0, "Invalid level %d", level);
-		goto err;
+
+	if (level < 0 && !offline) {
+		/* find level by idx */
+		ret = find_level_by_delta(dev, di->images[idx]->file, &level);
+		if (ret) {
+			ploop_log(0, "Can't find %s level by delta %s, "
+					"assuming offline replace",
+					dev, di->images[idx]->file);
+			offline = 1;
+		}
 	}
 
 	/* check a new image */
-	ret = ploop_check(file, CHECK_DETAILED, 1, 0, 0, NULL);
+	ret = ploop_check(file, CHECK_DETAILED | CHECK_READONLY, NULL);
 	if (ret)
-		return ret;
+		goto err;
 
 	/* check that images are identical */
-	oldfile = param->cur_file ? : di->images[level]->file;
+	oldfile = param->cur_file ? : di->images[idx]->file;
 	ret = check_deltas_same(file, oldfile);
 	if (ret)
-		return ret;
+		goto err;
+
+	if (keep_name && cant_rename(file, oldfile)) {
+		ret = SYSEXIT_RENAME;
+		goto err;
+	}
 
 	/* Write new dd.xml with changed image file */
 	get_disk_descriptor_fname(di, conf, sizeof(conf));
 	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
-	oldfile = di->images[level]->file;
-	di->images[level]->file = file;
+	tmp = di->images[idx]->file;
+	di->images[idx]->file = file;
 	ret = ploop_store_diskdescriptor(conf_tmp, di);
-	di->images[level]->file = oldfile;
+	di->images[idx]->file = tmp;
 	if (ret)
 		goto err;
 
 	/* Do replace */
-	ploop_log(0, "Replacing %s with %s", di->images[level]->file, file);
-	ret = replace_delta(dev, level, file);
-	if (ret)
-		goto err;
-
-	/* Put a new dd.xml */
-	if (rename(conf_tmp, conf)) {
-		ploop_err(errno, "Can't rename %s to %s", conf_tmp, conf);
-		ret = SYSEXIT_RENAME;
-		/* FIXME: how to rollback now? */
-		goto err;
+	ploop_log(0, "Replacing %s with %s (%s, level %d)", oldfile, file,
+			(offline) ? "offline" : "online", level);
+	if (!offline) {
+	       ret = replace_delta(dev, level, file);
+	       if (ret)
+		       goto err;
 	}
-	conf_tmp[0] = '\0'; /* prevent unlink() below */
 
-	/* Change image in di */
-	free(di->images[level]->file);
-	di->images[level]->file = file; /* malloc()ed by realpath */
-	file = NULL; /* prevent free(file) below */
+	if (keep_name) {
+		char tmp[PATH_MAX] = "";
+
+		ret = SYSEXIT_SYS;
+
+		if (!offline && st_nlink(file) < 2) {
+			/* If ploop is running, we can't just rename
+			 * the file if its st.st_nlink < 2 as it is used
+			 * by ploop and the kernel checks that the last
+			 * reference to the file is not removed.
+			 *
+			 * We need to create a hardlink to it,
+			 * rename the file, then remove the hardlink.
+			 */
+			snprintf(tmp, sizeof(tmp), "%s.XXXXXX", file);
+			if (mktemp(tmp)[0] == '\0') {
+				ploop_err(errno, "Can't mktemp(%s)", tmp);
+				goto undo_keep;
+			}
+			if (link(file, tmp)) {
+				ploop_err(errno, "Can't hardlink %s to %s",
+						tmp, file);
+				goto undo_keep;
+			}
+		}
+		if (rename(file, oldfile)) {
+			ploop_err(errno, "Can't rename %s to %s",
+					file, oldfile);
+			goto undo_keep;
+		}
+
+		ret = 0;
+
+undo_keep:
+		if (tmp[0] && unlink(tmp)) {
+			ploop_err(errno, "Can't delete %s", tmp);
+		}
+
+		if (ret && !offline) {
+			ploop_log(0, "Rollback: replacing %s with %s",
+					file, oldfile);
+			if (replace_delta(dev, level, oldfile)) {
+				/* Hmm. We can't roll back the replace, so
+				 * let's at least keep the dd.xml consistent
+				 * with the in-kernel ploop state.
+				 */
+				ploop_log(0, "Rollback replace failed, "
+						"saving new image name "
+						"to DiskDescriptor.xml");
+				ret = 0; /* FIXME: do we want error code? */
+				keep_name = 0;
+			}
+		}
+	}
+
+	if (!keep_name) {
+		/* Put a new dd.xml */
+		ret = rename(conf_tmp, conf);
+		conf_tmp[0] = '\0'; /* prevent unlink() below */
+		if (ret) {
+			ploop_err(errno, "Can't rename %s to %s",
+					conf_tmp, conf);
+			ret = SYSEXIT_RENAME;
+			/* FIXME: how to rollback now? */
+			goto err;
+		}
+		/* Change image in di */
+		free(di->images[idx]->file);
+		di->images[idx]->file = file; /* malloc()ed by realpath */
+		file = NULL; /* prevent free(file) below */
+	}
 
 	ret = 0;
 err:
@@ -1505,6 +1715,14 @@ static int create_ploop_dev(int minor)
 	return 0;
 }
 
+static int set_max_delta_size(int fd, __u64 size)
+{
+	/* Set max delta size for the last added (top) delta */
+	ploop_log(0, "Setting maximum delta size to %llu sec", size);
+
+	return ioctl_device(fd, PLOOP_IOC_MAX_DELTA_SIZE, &size);
+}
+
 /* NB: caller will take care about *lfd_p even if we fail */
 static int add_deltas(struct ploop_disk_images_data *di,
 		char **images, struct ploop_mount_param *param,
@@ -1568,6 +1786,11 @@ static int add_deltas(struct ploop_disk_images_data *di,
 		if (ret)
 			goto err1;
 	}
+
+	if (di != NULL && di->max_delta_size != 0 &&
+			(ret = set_max_delta_size(*lfd_p, di->max_delta_size)))
+		goto err1;
+
 	if (ioctl(*lfd_p, PLOOP_IOC_START, 0) < 0) {
 		ploop_err(errno, "PLOOP_IOC_START");
 		ret = SYSEXIT_DEVIOC;
@@ -1727,7 +1950,7 @@ static int check_mount_restrictions(char **images)
 
 	for (i = 0; images[i] != NULL; i++) {
 		if (stat(images[i], &st) < 0) {
-			ploop_err(errno, "Unable to statfs %s", images[i]);
+			ploop_err(errno, "Unable to stat %s", images[i]);
 			return -1;
 		}
 		/* device already checked */
@@ -1793,7 +2016,7 @@ int ploop_mount(struct ploop_disk_images_data *di, char **images,
 	if (ret)
 		goto err;
 
-	if (param->target != NULL) {
+	if (param->target != NULL || param->fsck) {
 		ret = ploop_mount_fs(param);
 		if (ret)
 			ploop_stop(lfd, param->device);
@@ -1845,7 +2068,7 @@ int mount_image(struct ploop_disk_images_data *di, struct ploop_mount_param *par
 
 	ret = ploop_mount(di, images, param, (di->mode == PLOOP_RAW_MODE));
 
-	free_images_list(images);
+	ploop_free_array(images);
 
 	return ret;
 }
@@ -1931,10 +2154,8 @@ static int ploop_umount_fs(const char *mnt, struct ploop_disk_images_data *di)
 		store_statfs_info(mnt, di->images[0]->file);
 	ploop_log(0, "Unmounting file system at %s", mnt);
 	ret = do_umount(mnt);
-	if (ret)
-		return ret;
 
-	return 0;
+	return ret;
 }
 
 int ploop_umount(const char *device, struct ploop_disk_images_data *di)
@@ -1944,7 +2165,7 @@ int ploop_umount(const char *device, struct ploop_disk_images_data *di)
 
 	if (!device) {
 		ploop_err(0, "ploop_umount: device is not specified");
-		return -1;
+		return SYSEXIT_SYS; /* internal error */
 	}
 
 	if (get_mount_dir(device, mnt, sizeof(mnt)) == 0) {
@@ -2137,14 +2358,24 @@ int ploop_grow_image(struct ploop_disk_images_data *di, off_t size)
 {
 	int ret;
 	char device[64];
+	char conf[PATH_MAX];
+	char conf_tmp[PATH_MAX];
 
 	if (ploop_lock_dd(di))
 		return SYSEXIT_LOCK;
 
+	// Update size in the DiskDescriptor.xml
+	di->size = size;
+	get_disk_descriptor_fname(di, conf, sizeof(conf));
+	snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
+	ret = ploop_store_diskdescriptor(conf_tmp, di);
+	if (ret)
+		goto err;
+
 	ret = ploop_find_dev_by_dd(di, device, sizeof(device));
 	if (ret == -1) {
 		ret = SYSEXIT_SYS;
-		goto err;
+		goto err_unlink;
 	}
 	if (ret == 0) {
 		ret = ploop_grow_device(device, size);
@@ -2156,14 +2387,14 @@ int ploop_grow_image(struct ploop_disk_images_data *di, off_t size)
 		if (i == -1) {
 			ploop_err(0, "Unable to find top delta file name");
 			ret = SYSEXIT_PARAM;
-			goto err;
+			goto err_unlink;
 		}
 
 		fname = find_image_by_guid(di, di->top_guid);
 		if (!fname) {
 			ploop_err(0, "Unable to find top delta file name");
 			ret = SYSEXIT_PARAM;
-			goto err;
+			goto err_unlink;
 		}
 
 		if (strcmp(di->snapshots[i]->parent_guid, NONE_UUID) == 0 &&
@@ -2173,6 +2404,17 @@ int ploop_grow_image(struct ploop_disk_images_data *di, off_t size)
 			ret = ploop_grow_delta_offline(fname, size);
 	}
 
+	if (ret)
+		goto err_unlink;
+
+	if (rename(conf_tmp, conf)) {
+		ploop_err(errno, "Can't rename %s to %s",
+				conf_tmp, conf);
+		ret = SYSEXIT_RENAME;
+	}
+
+err_unlink:
+	unlink(conf_tmp);
 err:
 	ploop_unlock_dd(di);
 
@@ -2478,6 +2720,7 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			struct dump2fs_data data = {};
 			__u64 available_balloon_size;
 			__u64 blocks;
+			__u64 reserved_blocks;
 
 			ret = dumpe2fs(part_device, &data);
 			if (ret)
@@ -2493,12 +2736,18 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 				}
 
 				new_balloon_size = blocks - new_fs_size;
+				/* exclude data accounted by ext4 by assumption that
+				 * overhead is inodes * inode size
+				 */
+				reserved_blocks = B2S(fs.f_files * 256);
+				new_balloon_size -= new_balloon_size > reserved_blocks ?
+							reserved_blocks : 0;
 				available_balloon_size = balloon_size + (fs.f_bfree * B2S(fs.f_bsize));
 				if (available_balloon_size < new_balloon_size) {
 					ploop_err(0, "Unable to change image size to %lu "
 							"sectors, minimal size is %llu",
 							(long)new_fs_size,
-							(blocks - available_balloon_size));
+							(blocks - available_balloon_size - reserved_blocks));
 					ret = SYSEXIT_PARAM;
 					goto err;
 				}
@@ -3178,3 +3427,321 @@ int ploop_get_spec(struct ploop_disk_images_data *di, struct ploop_spec *spec)
 
 	return ret;
 }
+
+int ploop_set_max_delta_size(struct ploop_disk_images_data *di, __u64 size)
+{
+	char conf[PATH_MAX];
+	char dev[64];
+	int fd = -1;
+	int rc;
+
+	if (ploop_lock_dd(di))
+		return SYSEXIT_LOCK;
+
+	rc = ploop_find_dev_by_dd(di, dev, sizeof(dev));
+	if (rc == -1) {
+		rc = SYSEXIT_SYS;
+		goto err;
+	} else if (rc == 0) {
+		fd = open(dev, O_RDONLY);
+		if (fd == -1) {
+			ploop_err(errno, "Can't open device %s", dev);
+			rc = SYSEXIT_DEVICE;
+			goto err;
+		}
+		rc = set_max_delta_size(fd, size);
+		if (rc)
+			goto err;
+	}
+
+	get_disk_descriptor_fname(di, conf, sizeof(conf));
+	di->max_delta_size = size;
+	rc = ploop_store_diskdescriptor(conf, di);
+err:
+	if (fd != -1)
+		close(fd);
+	ploop_unlock_dd(di);
+
+	return rc;
+
+}
+
+int copy_delta(const char *src, const char *dst)
+{
+	void *buf = NULL;
+	int sfd = -1, dfd = -1;
+	struct ploop_pvd_header *vh;
+	int version, cluster = DEF_CLUSTER;
+	struct stat st;
+	off_t i;
+	int ret = 0;
+
+	sfd = open(src, O_DIRECT | O_RDONLY);
+	if (sfd < 0) {
+		ploop_err(errno, "Can't open %s", src);
+		return SYSEXIT_OPEN;
+	}
+
+	dfd = open(dst, O_RDWR | O_CREAT | O_DIRECT | O_EXCL, 0600);
+	if (dfd < 0) {
+		ploop_err(errno, "Can't create %s", dst);
+		ret = SYSEXIT_CREAT;
+		goto out;
+	}
+
+	if (fstat(sfd, &st)) {
+		ploop_err(errno, "Can't fstat %s", src);
+		ret = SYSEXIT_OPEN;
+		goto out;
+	}
+
+	if (p_memalign(&buf, 4096, cluster)) {
+		ret = SYSEXIT_MALLOC;
+		goto out;
+	}
+
+	/* Read header */
+	ret = read_safe(sfd, buf, 4096, 0, "read PVD header");
+	if (ret)
+		goto out;
+	vh = buf;
+
+	/* Do some minor checks on header */
+	ret = SYSEXIT_PLOOPFMT;
+	version = ploop1_version(vh);
+	if (version == PLOOP_FMT_ERROR) {
+		ploop_err(0, "Wrong signature in image %s", src);
+		goto out;
+	}
+	if (vh->m_Type != PRL_IMAGE_COMPRESSED) {
+		ploop_err(0, "Wrong type in image %s", src);
+		goto out;
+	}
+	if (!is_valid_blocksize(vh->m_Sectors)) {
+		ploop_err(0, "Wrong cluster size %d in image %s",
+				vh->m_Sectors, src);
+		goto out;
+	}
+	if (vh->m_FirstBlockOffset % vh->m_Sectors != 0) {
+		ploop_err(0, "Wrong first block offset in image %s", src);
+		goto out;
+	}
+	if (vh->m_DiskInUse) {
+		ploop_err(0, "Image %s is in use", src);
+		ret = SYSEXIT_PLOOPINUSE;
+		goto out;
+	}
+
+	/* Does cluster size differs from default? */
+	if (cluster != S2B(vh->m_Sectors)) {
+		/* realloc */
+		cluster = S2B(vh->m_Sectors);
+		free(buf);
+		if (p_memalign(&buf, 4096, cluster)) {
+			ret = SYSEXIT_MALLOC;
+			goto out;
+		}
+	}
+
+	/* Check file size for sanity, should be X clusters */
+	if (st.st_size % cluster != 0) {
+		ploop_err(0, "Bad file size of image %s", src);
+		ret = SYSEXIT_PLOOPFMT;
+		goto out;
+	}
+
+	ploop_log(0, "Copying %lu MB delta %s to %s",
+			(unsigned long)(st.st_size >> 20), src, dst);
+
+	/* Preallocate disk space */
+	if (sys_fallocate(dfd, 0, 0, st.st_size) && errno != ENOTSUP) {
+		ploop_err(errno, "Can't fallocate(%s, %lu)",
+				dst, (unsigned long)st.st_size);
+		ret = SYSEXIT_FALLOCATE;
+		goto out;
+	}
+
+	/* Copy data */
+	for (i = 0; i < st.st_size; i+=cluster) {
+		ret = read_safe(sfd, buf, cluster, i, "read cluster");
+		if (ret)
+			goto out;
+		ret = write_safe(dfd, buf, cluster, i, "write cluster");
+		if (ret)
+			goto out;
+	}
+
+	if (fsync(dfd)) {
+		ploop_err(errno, "Failed to sync %s", dst);
+		ret = SYSEXIT_FSYNC;
+		goto out;
+	}
+
+	ret = 0;
+out:
+	if (buf)
+		free(buf);
+	if (sfd >= 0)
+		close(sfd);
+	if (dfd >= 0)
+		close(dfd);
+	if (ret && ret != SYSEXIT_CREAT)
+		unlink(dst);
+
+	return ret;
+}
+
+/*
+ * Find best blocksize for raw image.
+ * Select largest possible blocksize between 1M and 32K.
+ *
+ * Returns blocksize in sectors or -1 in case of error.
+ */
+static int select_best_blocksize(off_t size)
+{
+	int i;
+
+	for (i = 20; i >= 15; i--)
+		if (size % (1 << i) == 0)
+			return 1 << (i - PLOOP1_SECTOR_LOG);
+
+	return -1;
+}
+
+int ploop_restore_descriptor(const char *dir, char *delta_path, int raw, int blocksize)
+{
+	struct delta delta;
+	struct ploop_disk_images_data *di = NULL;
+	struct ploop_create_param param = {};
+	char ddxml[PATH_MAX];
+	char fname[PATH_MAX];
+	int ret;
+
+	if (strlen(dir) == 0)
+		return SYSEXIT_PARAM;
+
+	ret = snprintf(ddxml, sizeof(ddxml), "%s/" DISKDESCRIPTOR_XML, dir);
+	if (ret >= sizeof(ddxml)) {
+		ploop_err(0, "Output path is too long");
+		return SYSEXIT_PARAM;
+	}
+
+	if (raw) {
+		struct stat st;
+
+		param.mode = PLOOP_RAW_MODE;
+		if (stat(delta_path, &st)) {
+			ploop_err(errno, "stat %s", delta_path);
+			return SYSEXIT_OPEN;
+		}
+
+		if(blocksize) {
+			if (st.st_size % (blocksize * SECTOR_SIZE)) {
+				ploop_err(0, "Image size must be aligned "
+						"to the blocksize specified");
+				return SYSEXIT_PARAM;
+			}
+
+			param.blocksize = blocksize;
+		} else {
+			param.blocksize = select_best_blocksize(st.st_size);
+			if (param.blocksize < 0) {
+				ploop_err(0, "Image size must be aligned to 32K");
+				return SYSEXIT_PARAM;
+			}
+		}
+
+		param.size = st.st_size;
+		param.image = delta_path;
+	} else {
+		if (open_delta(&delta, delta_path, O_RDONLY, OD_ALLOW_DIRTY))
+			return SYSEXIT_OPEN;
+
+		param.size = delta.blocksize * delta.l2_size;
+		param.mode = PLOOP_EXPANDED_MODE;
+		param.image = delta_path;
+		param.blocksize = delta.blocksize;
+		param.fmt_version = delta.version;
+		close_delta(&delta);
+	}
+
+	ret = init_dd(&di, ddxml, &param);
+	if (ret)
+		return ret;
+
+	if (realpath(param.image, fname) == NULL) {
+		ploop_err(errno, "failed realpath(%s)", param.image);
+		ret = SYSEXIT_CREAT;
+		goto out;
+	}
+
+	ret = ploop_di_add_image(di, fname, TOPDELTA_UUID, NONE_UUID);
+	if (ret)
+		goto out;
+
+	ret = ploop_store_diskdescriptor(ddxml, di);
+	if (ret)
+		goto out;
+
+out:
+	ploop_close_dd(di);
+	return ret;
+}
+
+static const char *get_devs_str(char **devs, char *buf, int size)
+{
+	char **p;
+	char *sp = buf;
+	char *ep = buf + size;
+
+	for (p = devs; *p != NULL; p++) {
+		sp += snprintf(sp, ep - sp, "%s ", *p);
+		if (sp >= ep)
+			break;
+	}
+	return buf;
+}
+
+int check_snapshot_mount(struct ploop_disk_images_data *di,
+		const char *guid, const char *fname, int temp)
+{
+	int ret = 0;
+	char **devs, **p;
+	char buf[512];
+
+	ret = ploop_get_dev_by_delta(di->images[0]->file,
+			fname, NULL, &devs);
+	if (ret == 1)
+		return 0;
+	if (ret == -1)
+		return SYSEXIT_SYS;
+
+	/* There is one or more ploop devices found */
+	if (!temp) {
+		ploop_err(0, "Snapshot %s is busy by device(s): %s", guid,
+				get_devs_str(devs, buf, sizeof(buf)));
+		ret = SYSEXIT_EBUSY;
+		goto out;
+	}
+
+	/* Try to unmount temp snapshot(s) */
+	for (p = devs; *p != NULL; p++) {
+		if (!is_device_inuse(*p))
+			ret = ploop_umount(*p, NULL);
+		else
+			ret = SYSEXIT_EBUSY;
+
+		if (ret) {
+			/* print current device and remaining ones */
+			ploop_err(0, "Snapshot %s is busy by device(s): %s", guid,
+					get_devs_str(p, buf, sizeof(buf)));
+			goto out;
+		}
+	}
+
+	ret = 0;
+out:
+	ploop_free_array(devs);
+
+	return ret;
+}
