@@ -57,6 +57,9 @@
 #define CHECK_DROPINUSE	0x04
 #define CHECK_DETAILED	0x08
 #define CHECK_REPAIR_SPARSE	0x10
+#define CHECK_READONLY		0x20	/* do a read-only check */
+#define CHECK_TALKATIVE		0x40	/* be verbose, produce more output */
+#define CHECK_RAW		0x80	/* delta is in raw format */
 
 #define S2B(sec) ((off_t)(sec) << PLOOP1_SECTOR_LOG)
 #define B2S(sec) ((sec) >> PLOOP1_SECTOR_LOG)
@@ -69,17 +72,6 @@
 #define DEF_PATH_LIST	{ "/sbin", "/bin", "/usr/sbin", "/usr/bin", \
 			"/usr/local/sbin", "/usr/bin", NULL }
 
-struct delta_fops
-{
-	int		(*open)(char *pathname, int flags, mode_t mode);
-	int		(*close)(int fd);
-	int		(*pread)(int fd, void *buf, size_t count, off_t offset);
-	int		(*pwrite)(int fd, void *buf, size_t count, off_t offset);
-	int		(*fstat)(int fd, struct stat *buf);
-	int		(*fsync)(int fd);
-	int		(*update_size)(int fd, const char *pathname);
-};
-
 struct delta
 {
 	int    fd;
@@ -96,7 +88,7 @@ struct delta
 	__u32  blocksize;
 	int    version;	  /* ploop1 version */
 
-	struct delta_fops *fops;
+	void *reserved1;
 };
 
 struct delta_array
@@ -223,14 +215,14 @@ static inline off_t ploop_ioff_to_sec(__u32 iblk, __u32 blocksize, int version)
 int gen_uuid_pair(char *uuid1, int len1, char *uuid2, int len2);
 int find_delta_names(const char * device, int start_level, int end_level,
 			    char **names, char ** format);
-PL_EXT int find_level_by_delta(const char *device, const char *delta);
+PL_EXT int find_level_by_delta(const char *device, const char *delta, int *level);
 PL_EXT int ploop_get_attr(const char * device, const char * attr, int * res);
 int ploop_get_delta_attr(const char * device, int level, const char * attr, int * res);
+int ploop_get_delta_attr_str(const char *device, int level, const char *attr, char *out, int len);
 int ploop_get_size(const char * device, off_t * res);
 int dev_num2dev_start(const char *device, dev_t dev_num, __u32 *dev_start);
 int get_dev_by_name(const char *name, dev_t *dev);
 int get_dev_start(const char *path, __u32 *start);
-int ploop_get_top_level(int devfd, const char *devname, int *top);
 
 void init_delta_array(struct delta_array *);
 void deinit_delta_array(struct delta_array * p);
@@ -277,8 +269,7 @@ struct relocmap *relocmap_alloc(int n);
 struct ploop_relocblks_ctl;
 int relocmap2relocblks(struct relocmap *relocmap, int lvl, __u32 a_h, __u32 n_scanned,
 			struct ploop_relocblks_ctl **relocblks_pp);
-PL_EXT int ploop_check(char *img, int flags, int ro, int raw, int verbose,
-		__u32 *blocksize_p);
+PL_EXT int ploop_check(char *img, int flags, __u32 *blocksize_p);
 int check_deltas(struct ploop_disk_images_data *di, char **images,
 		int raw, __u32 *blocksize);
 PL_EXT int check_dd(struct ploop_disk_images_data *di, const char *uuid);
@@ -317,7 +308,7 @@ PL_EXT int ploop_mount(struct ploop_disk_images_data *di, char **images,
 PL_EXT int replace_delta(const char *device, int level, const char *image);
 PL_EXT int create_snapshot(const char *device, const char *delta, int syncfs);
 int get_list_size(char **list);
-PL_EXT void free_images_list(char **images);
+int normalize_image_name(const char *basedir, const char *image, char *out, int len);
 int PWRITE(struct delta * delta, void * buf, unsigned int size, off_t off);
 int PREAD(struct delta * delta, void *buf, unsigned int size, off_t off);
 PL_EXT int ploop_getdevice(int *minor);
@@ -326,6 +317,7 @@ int read_diskdescriptor(const char *fname, struct ploop_disk_images_data *di);
 void get_disk_descriptor_fname(struct ploop_disk_images_data *di, char *buf, int size);
 void get_disk_descriptor_lock_fname(struct ploop_disk_images_data *di, char *out, int size);
 int find_image_idx_by_guid(struct ploop_disk_images_data *di, const char *guid);
+int find_image_idx_by_file(struct ploop_disk_images_data *di, const char *file);
 int ploop_find_dev(const char *component_name, const char *image, char *out, int size);
 int ploop_find_dev_by_cn(struct ploop_disk_images_data *di, const char *component_name,
 		int check_state, char *out, int len);
@@ -342,6 +334,8 @@ int get_image_param(struct ploop_disk_images_data *di, const char *guid,
 int get_image_param_offline(struct ploop_disk_images_data *di, const char *guid,
                 off_t *size, __u32 *blocksize, int *version);
 PL_EXT char **make_images_list(struct ploop_disk_images_data *di, const char *guid, int reverse);
+char **make_images_list_by_guids(struct ploop_disk_images_data *di,
+		const char *guid, const char *guid2, int reverse);
 
 // manage struct ploop_disk_images_data
 int ploop_di_add_image(struct ploop_disk_images_data *di, const char *fname,
@@ -349,13 +343,18 @@ int ploop_di_add_image(struct ploop_disk_images_data *di, const char *fname,
 void ploop_di_set_temporary(struct ploop_disk_images_data *di, const char *guid);
 int ploop_di_remove_image(struct ploop_disk_images_data *di, const char *guid,
 		int renew_top_uuid, char **fname);
-int ploop_di_merge_image(struct ploop_disk_images_data *di, const char *guid, char **fname);
+int ploop_di_merge_image(struct ploop_disk_images_data *di, const char *guid);
+int ploop_di_can_merge_images(struct ploop_disk_images_data *di,
+		const char *ancestor_guid, const char *descendant_guid);
+int ploop_di_merge_images(struct ploop_disk_images_data *di,
+		const char *guid, int depth);
 void ploop_di_change_guid(struct ploop_disk_images_data *di, const char *guid, const char *new_guid);
 PL_EXT char *find_image_by_guid(struct ploop_disk_images_data *di, const char *guid);
 PL_EXT int find_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid);
 int ploop_add_image_entry(struct ploop_disk_images_data *di, const char *fname, const char *guid);
 int ploop_add_snapshot_entry(struct ploop_disk_images_data *di, const char *guid,
 		const char *parent_guid, int temporary);
+char *get_base_delta_uuid(struct ploop_disk_images_data *di);
 
 //balloon
 PL_EXT char *mntn2str(int mntn_type);
@@ -400,7 +399,7 @@ int drop_statfs_info(const char *image);
 int read_statfs_info(const char *image, struct ploop_info *info);
 int get_statfs_info(const char *mnt, struct ploop_info *info);
 int ploop_get_child_count_by_uuid(struct ploop_disk_images_data *di, const char *guid);
-int ploop_get_child_by_uuid(struct ploop_disk_images_data *di, const char *guid,  char **child_guid);
+const char* ploop_find_child_by_guid(struct ploop_disk_images_data *di, const char *guid);
 int ploop_fname_cmp(const char *p1, const char *p2);
 PL_EXT int is_valid_guid(const char *guid);
 PL_EXT int read_line(const char *path, char *nbuf, int len);
@@ -419,11 +418,13 @@ int print_output(int level, const char *cmd, const char *arg);
 int read_safe(int fd, void * buf, unsigned int size, off_t pos, char *msg);
 int write_safe(int fd, void * buf, unsigned int size, off_t pos, char *msg);
 const char *get_snap_str(int temporary);
+PL_EXT int ploop_restore_descriptor(const char *dir, char *delta_path, int raw, int blocksize);
+int is_device_inuse(const char *dev);
 // merge
 PL_EXT int get_delta_info(const char *device, struct merge_info *info);
 PL_EXT int merge_image(const char *device, int start_level, int end_level, int raw, int merge_top,
-		char **images);
-int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid, int merge_mode);
+		char **images, const char *new_delta);
+int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid, int merge_mode, const char *new_delta);
 int merge_temporary_snapshots(struct ploop_disk_images_data *di);
 
 PL_EXT int ploop_change_fmt_version(struct ploop_disk_images_data *di,
@@ -431,6 +432,11 @@ PL_EXT int ploop_change_fmt_version(struct ploop_disk_images_data *di,
 int ploop_get_dev_by_delta(const char *delta, const char *topdelta,
 		const char *component_name, char **out[]);
 int check_snapshot_mount(struct ploop_disk_images_data *di,
-		int temporary, const char *parent_fname,
-		const char *child_fname, const char *child_guid);
+		const char *guid, const char *fname, int temp);
+int create_image(const char *file, __u32 blocksize, off_t size_sec, int mode,
+		int version);
+int do_replace_delta(int devfd, int level, int imgfd, __u32 blocksize,
+		const char *image);
+int copy_delta(const char *src, const char *dst);
+
 #endif
